% TCS Alignment Toolbox
%
% Copyright (C) 2013, 2014
% Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
% AG Theoretical Computer Science
% Centre of Excellence Cognitive Interaction Technology (CITEC)
% University of Bielefeld
% 
% This program is free software: you can redistribute it and/or modify
% it under the terms of the GNU Affero General Public License as
% published by the Free Software Foundation, either version 3 of the
% License, or (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU Affero General Public License for more details.
% 
% You should have received a copy of the GNU Affero General Public License
% along with this program.  If not, see <http://www.gnu.org/licenses/>.


% --- Robot Angles Example ---
%
% In this example we compare two artificial sequences of angles of a
% three-joint robotic arm. This is an example for manual sequence
% construction and is only recommended for more advanced users. Otherwise
% please refer to one of the other examples.

%% Preamble
%
% This includes the necessary classes from the TCSAlignmentToolbox
% .jar-file.

clear java;
clear;
clc;

javaaddpath TCSAlignmentToolbox-1.3.1.jar;
import de.unibi.citec.tcs.alignment.sequence.*;
import de.unibi.citec.tcs.alignment.*;
import de.unibi.citec.tcs.alignment.comparators.*;

%% 1.) Create keyword specifications for three angles, one timestamp,
% one movement type and a comment.

keywordSpecs = javaArray('de.unibi.citec.tcs.alignment.sequence.KeywordSpecification',6);
keywordSpecs(1) = VectorialKeywordSpecification(1,'alpha');
keywordSpecs(2) = VectorialKeywordSpecification(1,'beta');
keywordSpecs(3) = VectorialKeywordSpecification(1,'gamma');
keywordSpecs(4) = VectorialKeywordSpecification(1,'timestamp');
movement_types = Alphabet('pause|movement');
keywordSpecs(5) = SymbolicKeywordSpecification(movement_types,'movement_type');
keywordSpecs(6) = StringKeywordSpecification('comment');

%% 2.) Create NodeSpecification.

nodeSpec = NodeSpecification(keywordSpecs);

%% 3.) Build two sequences for that NodeSpecification.
%
% For these full blown sequences this might be a tedious task to do by hand
% so it is recommended to program a conversion from a more convenient
% format of your choice. For some easy cases we provide pre-defined
% conversions (have a look at the sine wave and the string edit distance
% example).

a = Sequence(nodeSpec);

n = Node(a);
n.setValue('alpha', VectorialValue(30));
n.setValue('beta', VectorialValue(10));
n.setValue('gamma', VectorialValue(10));
n.setValue('timestamp', VectorialValue(0));
n.setValue('movement_type', SymbolicValue(movement_types,'pause'));
n.setValue('comment', StringValue('Begin of pause before first movement'));
a.getNodes().add(n);
n = Node(a);
n.setValue('alpha', VectorialValue(30));
n.setValue('beta', VectorialValue(10));
n.setValue('gamma', VectorialValue(10));
n.setValue('timestamp', VectorialValue(1));
n.setValue('movement_type', SymbolicValue(movement_types,'pause'));
a.getNodes().add(n);
n = Node(a);
n.setValue('alpha', VectorialValue(40));
n.setValue('beta', VectorialValue(10));
n.setValue('gamma', VectorialValue(10));
n.setValue('timestamp', VectorialValue(2.2));
n.setValue('movement_type', SymbolicValue(movement_types,'movement'));
n.setValue('comment', StringValue('Begin of movement. Quite rapidly at the start.'));
a.getNodes().add(n);
n = Node(a);
n.setValue('alpha', VectorialValue(45));
n.setValue('beta', VectorialValue(10));
n.setValue('gamma', VectorialValue(10));
n.setValue('timestamp', VectorialValue(3));
n.setValue('movement_type', SymbolicValue(movement_types,'movement'));
a.getNodes().add(n);
n = Node(a);
n.setValue('alpha', VectorialValue(50));
n.setValue('beta', VectorialValue(10));
n.setValue('gamma', VectorialValue(10));
n.setValue('timestamp', VectorialValue(4));
n.setValue('movement_type', SymbolicValue(movement_types,'movement'));
n.setValue('comment', StringValue('End of movement. End of sequence.'));
a.getNodes().add(n);


b = Sequence(nodeSpec);

n = Node(b);
n.setValue('alpha', VectorialValue(30));
n.setValue('beta', VectorialValue(10));
n.setValue('gamma', VectorialValue(10));
n.setValue('timestamp', VectorialValue(0));
n.setValue('movement_type', SymbolicValue(movement_types,'pause'));
n.setValue('comment', StringValue('Begin of pause before first movement'));
b.getNodes().add(n);
n = Node(b);
n.setValue('alpha', VectorialValue(32));
n.setValue('beta', VectorialValue(12));
n.setValue('gamma', VectorialValue(10));
n.setValue('timestamp', VectorialValue(0.8));
n.setValue('movement_type', SymbolicValue(movement_types,'movement'));
n.setValue('comment', StringValue('Begin of movement.'));
b.getNodes().add(n);
n = Node(b);
n.setValue('alpha', VectorialValue(36));
n.setValue('beta', VectorialValue(7));
n.setValue('gamma', VectorialValue(10));
n.setValue('timestamp', VectorialValue(2));
n.setValue('movement_type', SymbolicValue(movement_types,'movement'));
b.getNodes().add(n);
n = Node(b);
n.setValue('alpha', VectorialValue(42));
n.setValue('beta', VectorialValue(16));
n.setValue('gamma', VectorialValue(10));
n.setValue('timestamp', VectorialValue(3.3));
n.setValue('movement_type', SymbolicValue(movement_types,'movement'));
b.getNodes().add(n);
n = Node(b);
n.setValue('alpha', VectorialValue(50));
n.setValue('beta', VectorialValue(10));
n.setValue('gamma', VectorialValue(10));
n.setValue('timestamp', VectorialValue(4));
n.setValue('movement_type', SymbolicValue(movement_types,'movement'));
n.setValue('comment', StringValue('End of movement. End of sequence.'));
b.getNodes().add(n);

%% show both sequences.

display('a = ');
a.toString
display('b = ');
b.toString

%% 4.) Create alignment specification
% we want to use all keywords except for the comment keyword.
% As comparators we use the L1-Norm (abs) for the angles and the time stamp
% and a TrivialEditComparator for the movement_type.
% Please have a look in the javadoc for further information on the
% comparators.

align_keywords = javaArray('java.lang.String',5);
align_comparators = javaArray('de.unibi.citec.tcs.alignment.comparators.Comparator',5);
for k=1:5
    align_keywords(k) = keywordSpecs(k).getKeyword();
end

for k=1:4
    align_comparators(k) = L1NormComparator(1);
end

align_comparators(5) = TrivialEditComparator();

%% 5.) our weighting should assign emphasis on the alpha angle and the movement
% type, while the timestamp does not interest us as much.

weighting = [0.3,0.15,0.15,0.1,0.3];

alignSpec = AlignmentSpecification(nodeSpec,align_keywords,align_comparators,weighting);

%% visualize the alignment specification with the given defaults

display('Alignment specification = ');

alignSpec.toString()

%% 6.) Specify the alignment algorithm

algo = StrictAlignmentFullAlgorithm(alignSpec);

%% 8.) calculate alignment

alignResult = algo.calculateAlignment(a,b);

%% show the result

display('Alignment result = ');

alignResult.toString

%% 9.) calculate derivative for the weights. This corresponds to
% statistics over the distance for the comparators: In the strict case we
% just sum up over all distances for a given comparator along the optimal
% alignment path. In our example the gamma angle is consistent for both
% paths, leading to a derivative of 0. Alpha and Beta have some deviation
% as do timestamp and movement type. Note that this derivative does _not_
% consider the current weighting, of course.
%
% In a simple learning approach you could now choose to emphasize certain
% comparators more (or less) if you want to ensure a bigger (lower)
% distance (please refer to the sine waves example for a closer look on that).

weight_derivative = alignResult.calculateWeightDerivative()

% You could calculate derivatives also for the comparators, but we did not
% use a derivable comparator in this example.