/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.comparators.SkipComparator;
import de.unibi.citec.tcs.alignment.sequence.Node;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.Value;
import java.util.ArrayList;
import java.util.EnumMap;
import java.util.Stack;

/**
 * This is an implementation of the local alignment algorithm scheme
 * combining actually two approaches: First the algorithm of Gotoh et al.
 * for gap extensions (meaning, that we allow to skip longer regions in the
 * middle of an alignment at low cost) and second the traditional local
 * alignment in the sense that we allow skips at the beginning and end of an
 * alignment. The task here is to identify those regions of the two input
 * sequence that match the best and skip the rest of the sequences at low cost.
 *
 * A special thanks goes to Prof. Robert Giegerich, Dr. Stefan Janssen and
 * Benedikt Löwes at Bielefeld University for teaching me Advanced Dynamic
 * Programming. Otherwise this implementation here would not have been possible.
 *
 * The abstract algorithm that inspires this implementation is described in
 * Dr. Georg Sauthoffs "Bellman's Gap" GAP-Language in the file
 * bellmans_gap/local_alignment.gap.
 * 
 * TODO: THIS IS NOT YET IMPLEMENTED!
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public abstract class AbstractLocalAlignmentAlgorithm<R> implements LocalAlignmentAlgorithm<R> {

	private final Class<R> resultClass;
	private final AlignmentSpecification alignmentSpecification;
	private double weightThreshold = 0;

	public AbstractLocalAlignmentAlgorithm(Class<R> resultClass, AlignmentSpecification alignmentSpecification) {
		this.resultClass = resultClass;
		this.alignmentSpecification = alignmentSpecification;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public AlignmentSpecification getSpecification() {
		return alignmentSpecification;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public Class<R> getResultClass() {
		return resultClass;
	}

	/**
	 * Set a weight threshold (between 0 and 1) that determines which keywords
	 * should be ignored during calculation because their weight is negligible.
	 *
	 * The default value is 0.
	 *
	 * @param weightThreshold a weight threshold (between 0 and 1)
	 */
	public void setWeightThreshold(double weightThreshold) {
		if (weightThreshold < 0 || weightThreshold > 1) {
			throw new RuntimeException("A weight threshold has to be between 0 and 1!");
		}
		this.weightThreshold = weightThreshold;
	}

	/**
	 *
	 * @return The current weight threshold (0 per default).
	 */
	public double getWeightThreshold() {
		return weightThreshold;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public R calculateAlignment(Sequence a, Sequence b) {

		//check validity
		if (a.getNodeSpecficiation() != alignmentSpecification.getNodeSpecification()
				&& !a.getNodeSpecficiation().equals(alignmentSpecification.getNodeSpecification())) {
			throw new RuntimeException(
					"The first input sequence has an unexpected node specification!");
		}
		if (a.getNodeSpecficiation() != b.getNodeSpecficiation()
				&& !a.getNodeSpecficiation().equals(b.getNodeSpecficiation())) {
			throw new RuntimeException(
					"The node specifications of both input sequences to not match!");
		}
		//check validity of comparators.
		for (int k = 0; k < alignmentSpecification.size(); k++) {
			if (!(alignmentSpecification.getComparator(k) instanceof SkipComparator)) {
				throw new RuntimeException("The comparator for keyword "
						+ alignmentSpecification.getKeyword(k) + " does not support skips!");
			}
		}

		//identify the subset of comparators that have an above threshold weighting.
		final ArrayList<Integer> relevantIndices = new ArrayList<Integer>();
		for (int k = 0; k < alignmentSpecification.size(); k++) {
			if (alignmentSpecification.getWeighting()[k] > weightThreshold) {
				relevantIndices.add(k);
			}
		}

		//pre-calculation caching.
		final SkipComparator[] comparators = new SkipComparator[relevantIndices.size()];
		final double[] weights = new double[relevantIndices.size()];
		final int[] originalIndices = new int[relevantIndices.size()];
		for (int k = 0; k < comparators.length; k++) {
			comparators[k] = (SkipComparator) alignmentSpecification.getComparator(relevantIndices.get(k));
			weights[k] = alignmentSpecification.getWeighting()[relevantIndices.get(k)];
			originalIndices[k] = alignmentSpecification.getOriginalIndex(relevantIndices.get(k));
		}
		final int M = a.getNodes().size();
		final int N = b.getNodes().size();

		/*
		 * Initialize the dynamic programming tables for each recurrence.
		 */
		final EnumMap<Recurrence, double[][]> dp_tables = new EnumMap<>(Recurrence.class);
//		final EnumMap<Recurrence, boolean[][]> computed = new EnumMap<>(Recurrence.class);
		for (final Recurrence rec : Recurrence.values()) {
			dp_tables.put(rec, new double[M + 1][N + 1]);
//			computed.put(rec, new boolean[M + 1][N + 1]);
		}

		Node left;
		Node right;
		/*
		 * Initialize the cost for skipping both sequences entirely.
		 */
		for (int i = 1; i <= M; i++) {
			left = a.getNodes().get(i - 1);
			double local = 0;
			for (int k = 0; k < comparators.length; k++) {
				local += weights[k] * comparators[k].skipDelete(
						left.getValue(originalIndices[k]));
			}
			
		}

		/*
		 * Note that we need a lot of recurrences for this kind of local
		 * alignment, at least to do it efficiently.
		 */
//		final Stack<MatrixCoordinate> recStack = new Stack<>();
//		recStack.push(new MatrixCoordinate(0, 0, Recurrence.SKIPDEL_START));
//
//		while (!recStack.empty()) {
//			final MatrixCoordinate c = recStack.pop();
//			if (computed.get(c.rec)[c.i][c.j]) {
//				continue;
//			}
//			switch (c.rec) {
//				case SKIPDEL_START: {
//					//compute score for skip_ins option starting here.
//					if (!computed.get(Recurrence.SKIPINS_START)[c.i][c.j]) {
//						recStack.push(c);
//						recStack.push(new MatrixCoordinate(c.i, c.j,
//								Recurrence.SKIPINS_START));
//						continue;
//					}
//					//compute score for continuing with skip-deletions.
//					if (c.i < M && !computed.get(Recurrence.SKIPDEL_START)[c.i + 1][c.j]) {
//						recStack.push(c);
//						recStack.push(new MatrixCoordinate(c.i + 1, c.j,
//								Recurrence.SKIPDEL_START));
//						continue;
//					}
//					final double skipins_cost
//							= dp_tables.get(Recurrence.SKIPINS_START)[c.i][c.j];
//					final double[][] table = dp_tables.get(Recurrence.SKIPDEL_START);
//					if (c.i < M) {
//						left = a.getNodes().get(c.i);
//						double local = 0;
//						for (int k = 0; k < comparators.length; k++) {
//							local += weights[k] * comparators[k].skipDelete(
//									left.getValue(originalIndices[k]));
//						}
//						table[c.i][c.j] = choice(skipins_cost,
//								table[c.i + 1][c.j] + local);
//					} else {
//						table[c.i][c.j] = skipins_cost;
//					}
//					computed.get(Recurrence.SKIPDEL_START)[c.i][c.j] = true;
//				}
//				break;
//				case SKIPINS_START: {
//					//it is possible to complete the alignment using only
//					//skip-deletions and skip-insertions.
//					if (c.i == M && c.j == N) {
//						computed.get(Recurrence.SKIPINS_START)[M][N] = true;
//						continue;
//					}
//					//compute the score for an interesting section of the alignment
//					//starting here.
//					if (c.i < M && c.j < N && !computed.get(Recurrence.ALI)[c.i + 1][c.j + 1]) {
//						recStack.push(c);
//						recStack.push(new MatrixCoordinate(c.i + 1, c.j + 1,
//								Recurrence.ALI));
//						continue;
//					}
//					//compute score for continuing with skip-insertions.
//					if (c.j < N && !computed.get(Recurrence.SKIPINS_START)[c.i][c.j + 1]) {
//						recStack.push(c);
//						recStack.push(new MatrixCoordinate(c.i, c.j + 1,
//								Recurrence.SKIPINS_START));
//						continue;
//					}
//					final double[][] table = dp_tables.get(Recurrence.SKIPINS_START);
//					if (c.j < N) {
//						if (c.i < M) {
//							//if we can use a replacement, we try.
//							left = a.getNodes().get(c.i);
//							right = b.getNodes().get(c.j);
//							double repLocal = 0;
//							for (int k = 0; k < comparators.length; k++) {
//								repLocal += weights[k] * comparators[k].compare(
//										left.getValue(originalIndices[k]),
//										right.getValue(originalIndices[k]));
//							}
//							double local = 0;
//							for (int k = 0; k < comparators.length; k++) {
//								local += weights[k] * comparators[k].skipInsert(
//										right.getValue(originalIndices[k]));
//							}
//							table[c.i][c.j] = choice(
//									dp_tables.get(Recurrence.ALI)[c.i + 1][c.j + 1] + repLocal,
//									table[c.i][c.j + 1] + local);
//						} else {
//							//otherwise we just skip.
//							right = b.getNodes().get(c.j);
//							double local = 0;
//							for (int k = 0; k < comparators.length; k++) {
//								local += weights[k] * comparators[k].skipInsert(
//										right.getValue(originalIndices[k]));
//							}
//							table[c.i][c.j] = table[c.i][c.j + 1] + local;
//						}
//					} else {
//						//This is a forbidden state. It is not allowed that
//						//we are in the SKIPINS_START recurrence but can do
//						//no more skips.
//						table[c.i][c.j] = Double.POSITIVE_INFINITY;
//					}
//					computed.get(Recurrence.SKIPINS_START)[c.i][c.j] = true;
//				}
//				break;
//				case ALI:
//					//TODO: Implement.
//					computed.get(Recurrence.ALI)[c.i][c.j] = true;
//			}
//		}

		return transformToResult(dp_tables, a, b);
	}

//	private static class MatrixCoordinate {
//
//		final int i;
//		final int j;
//		final Recurrence rec;
//
//		public MatrixCoordinate(int i, int j, Recurrence rec) {
//			this.i = i;
//			this.j = j;
//			this.rec = rec;
//		}
//
//	}

	public static enum Recurrence {

		SKIPDEL_START,
		SKIPINS_START,
		ALI,
		DEL,
		INS,
		SKIPINS_MIDDLE,
		SKIPDEL_MIDDLE,
		SKIPINS_END,
		SKIPDEL_END;
	}

	/**
	 * This should implement the choice function as the term is used
	 * in the Bellman's Gap context. In the strict case this is a minimum,
	 * in the soft case a softmin applied on the input.
	 *
	 * @param choices the costs of all choices.
	 * @return a cost chosen from the given input.
	 */
	public abstract double choice(double... choices);

	/**
	 * This method should not be called from outside!
	 *
	 * The subclass uses this method to transform the alignment matrix and the
	 * input sequences to the actual alignment result.
	 *
	 * @param dp_tables the dynamic programming matrices used during
	 * calculation.
	 * @param a the first sequence.
	 * @param b th second sequence.
	 *
	 * @return the actual alignment result.
	 */
	public abstract R transformToResult(
			final EnumMap<Recurrence, double[][]> dp_tables,
			final Sequence a,
			final Sequence b);

	/**
	 * This calculates the non-weighted costs for skip-deleting the single
	 * values in node a.
	 *
	 * @param a The node from the left sequence.
	 * @param specification The AlignmentSpecification that shall be used (and
	 * is compatible with the nodes content).
	 *
	 * @return the non-weighted costs for skip-deleting the single values in
	 * node a.
	 */
	public static double[] calculateSkipDeletionCosts(final Node a,
			final AlignmentSpecification specification) {
		final double[] costs = new double[specification.size()];
		for (int k = 0; k < costs.length; k++) {
			final SkipComparator comp = (SkipComparator) specification.
					getComparator(k);
			costs[k] = comp.skipDelete(a.getValue(specification.getOriginalIndex(k)));
		}
		return costs;
	}

	/**
	 * This calculates the non-weighted costs for skip-inserting the single
	 * values of node b.
	 *
	 * @param b The node from the right sequence.
	 * @param specification The AlignmentSpecification that shall be used (and
	 * is compatible with the nodes content).
	 *
	 * @return the non-weighted costs for skip-inserting the single values of
	 * node b.
	 */
	public static double[] calculateInsertionCosts(final Node b,
			final AlignmentSpecification specification) {
		double[] costs = new double[specification.size()];
		for (int i = 0; i < costs.length; i++) {
			final SkipComparator comp = (SkipComparator) specification.
					getComparator(i);
			costs[i] = comp.skipInsert(b.getValue(specification.getOriginalIndex(i)));
		}
		return costs;
	}

}
