/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.sequence.Sequence;

/**
 * This is an implementation of the Needleman-Wunsch-Algorithm for sequence
 * alignment. It returns just the accumulated, normalized score. In each step
 * the score is 1 at max and in the worst case m+n steps are needed. Thus the
 * normalization is done by dividing by m+n.
 *
 * @author Benjamin Paassen - bpaassen(at)techfak.uni-bielefeld.de
 */
public class StrictAlignmentScoreAlgorithm
        extends AbstractGapAlignmentAlgorithm<StrictAlignmentScoreAlgorithm.ScoreEntry, Double> {

    /**
     * This sets up an AlignmentAlgorithm instance according to the given
     * specification.
     *
     * @param alignmentSpecification an AlignmentSpecification.
     */
    public StrictAlignmentScoreAlgorithm(AlignmentSpecification alignmentSpecification) {
        super(alignmentSpecification, StrictAlignmentScoreAlgorithm.ScoreEntry.class, Double.class);
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public ScoreEntry createInitial() {
        return new ScoreEntry(0);
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public ScoreEntry createDelInitial(ScoreEntry delOld, int i, double delLocal) {
        return new ScoreEntry(delOld.getScore() + delLocal);
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public ScoreEntry createInsInitial(ScoreEntry insOld, int j, double insLocal) {
        return new ScoreEntry(insOld.getScore() + insLocal);
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public ScoreEntry createNewEntry(ScoreEntry delOld, ScoreEntry insOld,
            ScoreEntry repOld,
            int i, int j,
            double delLocal, double insLocal, double repLocal) {
        final double delTotal = delOld.getScore() + delLocal;
        final double insTotal = insOld.getScore() + insLocal;
        final double repTotal = repOld.getScore() + repLocal;
        if (delTotal < insTotal) {
            if (delTotal < repTotal) {
                return new ScoreEntry(delTotal);
            } else {
                return new ScoreEntry(repTotal);
            }
        } else {
            if (insTotal < repTotal) {
                return new ScoreEntry(insTotal);
            } else {
                return new ScoreEntry(repTotal);
            }
        }
    }

    /**
     * {@inheritDoc }
     */
    @Override
    public Double transformToResult(ScoreEntry[][] alignmentMatrix,
            Sequence a, Sequence b) {
        final double rawScore = alignmentMatrix[a.getNodes().size()][b.getNodes().size()].
                getScore();
        /*
         * In each step of the alignment we had a distance of 1 at most. Thus we can normalize
         * the score to the range [0,1] if we just divide the raw score by the worst case, which
         * would be to delete a entirely and insert b entirely => worst case score = m+n.
         */
        return rawScore / ((double) (a.getNodes().size() + b.getNodes().size()));
    }

    public static class ScoreEntry implements AlignmentMatrixEntry {

        private final double totalScore;

        public ScoreEntry(double totalScore) {
            this.totalScore = totalScore;
        }

        public double getScore() {
            return totalScore;
        }
    }
}
