/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment.comparators;

import de.unibi.citec.tcs.alignment.sequence.StringValue;
import de.unibi.citec.tcs.alignment.sequence.ValueType;

/**
 * This compares strings using statistics over the number of letters inside.
 * This is more of a toy comparator.
 *
 * Please refer to the "distance" method below for more information.
 *
 * For gaps and skips this comparison method is extended by the
 * ComparisonBasedSkipExtendedComparator.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class CharStatComparator
		extends ComparisonBasedSkipExtendedComparator<StringValue> {

	public CharStatComparator() {
		super(new StringValue(""), new StringValue(""),
				new StringValue(""), new StringValue(""));
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double compare(StringValue a, StringValue b) {
		if (a == null && b == null) {
			return 0;
		}
		if (a == null && b != null) {
			return insert(b);
		}
		if (a != null && b == null) {
			return delete(a);
		}
		return distance(a.getString(), b.getString());
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public ValueType getType() {
		return ValueType.STRING;
	}

	/**
	 * This calculates the distance between the two strings a and b calculating
	 * for each alphanumeric character (a-z, 0-9) the number of times it occured
	 * in both strings at taking the difference of both. The sum over the
	 * absolute value of those differences is our basic metric. We then
	 * normalize by dividing by |a|+|b| (the input string lengths) to ensure
	 * that the resulting value is in the interval [0,1].
	 *
	 * You could also reformulate this as the L1-Norm of the vector of
	 * character-occurences normalized by the string length.
	 *
	 * @param a the first input string.
	 * @param b the second input string.
	 * @return a distance between a and b in the interval [0,1], where 0 means
	 * perfect match and 1 absolute dissimilarity.
	 */
	public static double distance(final String a, final String b) {
		//handle the trivial case that both strings are empty.
		if (a.isEmpty() && b.isEmpty()) {
			return 0;
		}
		//for our statistics we only consider 0-9 and a-z
		final int[] vector = new int[36];
		int code;
		//count the letters in string a.
		for (final char c : a.toCharArray()) {
			code = (int) Character.toUpperCase(c);
			if (code >= 48 && code <= 57) {
				vector[code - 48]++;
			} else if (code >= 65 && code <= 90) {
				vector[10 + (code - 65)]++;
			}
		}

		//subtract the letters in string b.
		for (final char c : b.toCharArray()) {
			code = (int) Character.toUpperCase(c);
			if (code >= 48 && code <= 57) {
				vector[code - 48]--;
			} else if (code >= 65 && code <= 90) {
				vector[10 + (code - 65)]--;
			}
		}

		//now sum over the vector and normalize.
		int sum = 0;
		for (int c = 0; c < vector.length; c++) {
			sum += Math.abs(vector[c]);
		}
		//normalize by length (how many percent of the letters were different?)
		final double normalizedDistance = ((double) sum) / ((double) (a.length() + b.length()));
		return normalizedDistance;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double comparisonDerivative(int paramIdx, StringValue a, StringValue b) {
		//this has no further parameters.
		return 0;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public SparseLocalDerivative sparseComparisonDerivative(StringValue a, StringValue b) {
		//this has no further parameters.
		return new SparseLocalDerivative.EmptySparseLocalEntry();
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int getNumberOfComparisonParameters() {
		//this has no further parameters.
		return 0;
	}
}
