/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment.comparators;

import de.unibi.citec.tcs.alignment.OperationType;
import de.unibi.citec.tcs.alignment.comparators.SparseLocalDerivative.SparseDeriativeEntry;
import de.unibi.citec.tcs.alignment.sequence.Value;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Objects;

/**
 * This is an abstract class allowing to define a dynamic method to
 * extend a simple Comparator to a SkipComparator.
 *
 * This provides a switch to choose, whether the dynamic method or the static
 * method of the SkipExtendedComparator abstract class should be used.
 *
 * The Comparison based method means that "compare" is called with a constant
 * comparison value to compute deletion, insertion, skipdeletion and
 * skipinsertion costs. It is also possible to set a factor (0.5 per default)
 * that makes skips cheaper than gaps.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 * @param <V> the value type this Comparator is defined on.
 */
public abstract class ComparisonBasedSkipExtendedComparator<V extends Value>
		extends SkipExtendedComparator<V> {

	private boolean dynamic = false;
	private V deleteComparison;
	private V insertComparison;
	private V skipDeleteComparison;
	private V skipInsertComparison;
	private double skipFactor = 0.5;

	public ComparisonBasedSkipExtendedComparator(V deleteComparison,
			V insertComparison, V skipDeleteComparison, V skipInsertComparison) {
		this.deleteComparison = deleteComparison;
		this.insertComparison = insertComparison;
		this.skipDeleteComparison = skipDeleteComparison;
		this.skipInsertComparison = skipInsertComparison;
	}

	/**
	 * Returns true if and only if this comparator will use a dynamic
	 * method to determine deletion, insertion and skip costs.
	 *
	 * @return true if and only if this comparator will use a dynamic
	 * method to determine deletion, insertion and skip costs.
	 */
	public boolean isDynamic() {
		return dynamic;
	}

	/**
	 * Set to true if and only if this comparator shall use a dynamic
	 * method to determine deletion, insertion and skip costs.
	 *
	 * @param dynamic true if and only if this comparator shall use a dynamic
	 * method to determine deletion, insertion and skip costs.
	 */
	public void setDynamic(boolean dynamic) {
		this.dynamic = dynamic;
	}

	/**
	 * Returns the cost factor for skipDeletions and skipInsertions for
	 * a dynamic comparison. E.g.: If value a is skip-deleted, skip_del is set
	 * as comparison for skipDeletion and c as skipFactor, than it will hold:
	 *
	 * skipDelete(a) = c * compare(a,s)
	 *
	 * @return the cost factor for skipDeletions and skipInsertions for
	 * a dynamic comparison.
	 */
	public double getSkipFactor() {
		return skipFactor;
	}

	/**
	 * Sets the cost factor for skipDeletions and skipInsertions for
	 * a dynamic comparison. E.g.: If value a is skip-deleted, skip_del is set
	 * as comparison for skipDeletion and c as skipFactor, than it will hold:
	 *
	 * skipDelete(a) = c * compare(a,s)
	 *
	 * @param skipFactor the cost factor for skipDeletions and skipInsertions
	 * for a dynamic comparison.
	 */
	public void setSkipFactor(double skipFactor) {
		this.skipFactor = skipFactor;
	}

	public V getDeleteComparison() {
		return deleteComparison;
	}

	public void setDeleteComparison(V deleteComparison) {
		this.deleteComparison = deleteComparison;
	}

	public V getInsertComparison() {
		return insertComparison;
	}

	public void setInsertComparison(V insertComparison) {
		this.insertComparison = insertComparison;
	}

	public V getSkipDeleteComparison() {
		return skipDeleteComparison;
	}

	public void setSkipDeleteComparison(V skipDeleteComparison) {
		this.skipDeleteComparison = skipDeleteComparison;
	}

	public V getSkipInsertComparison() {
		return skipInsertComparison;
	}

	public void setSkipInsertComparison(V skipInsertComparison) {
		this.skipInsertComparison = skipInsertComparison;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double delete(V a) {
		if (dynamic) {
			return compare(a, deleteComparison);
		} else {
			return super.delete(a);
		}
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double insert(V b) {
		if (dynamic) {
			return compare(insertComparison, b);
		} else {
			return super.insert(b);
		}
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double skipDelete(V a) {
		if (dynamic) {
			return skipFactor * compare(a, skipDeleteComparison);
		} else {
			return super.skipDelete(a);
		}
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double skipInsert(V b) {
		if (dynamic) {
			return skipFactor * compare(skipInsertComparison, b);
		} else {
			return super.skipInsert(b);
		}
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double calculateLocalDerivative(int paramIdx, V a, V b, OperationType type) {
		if (dynamic) {
			if (paramIdx < 4) {
				return 0;
			}
			switch (type) {
				case DELETION:
					return comparisonDerivative(paramIdx, a, deleteComparison);
				case INSERTION:
					return comparisonDerivative(paramIdx, insertComparison, b);
				case SKIPDELETION:
					return skipFactor * comparisonDerivative(paramIdx, a, skipDeleteComparison);
				case SKIPINSERTION:
					return skipFactor * comparisonDerivative(paramIdx, skipInsertComparison, b);
				case REPLACEMENT:
					return comparisonDerivative(paramIdx, a, b);
				default:
					throw new UnsupportedOperationException("Unknown operation type: " + type);
			}
		} else {
			return super.calculateLocalDerivative(paramIdx, a, b, type);
		}
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public SparseLocalDerivative calculateSparseLocalDerivative(V a, V b, OperationType type) {
		if (dynamic) {
			switch (type) {
				case DELETION:
					return sparseComparisonDerivative(a, deleteComparison);
				case INSERTION:
					return sparseComparisonDerivative(insertComparison, b);
				case SKIPDELETION:
					return multiplyWithSkipFactor(
							sparseComparisonDerivative(a, skipDeleteComparison));
				case SKIPINSERTION:
					return multiplyWithSkipFactor(
							sparseComparisonDerivative(skipInsertComparison, b));
				case REPLACEMENT:
					return sparseComparisonDerivative(a, b);
				default:
					throw new UnsupportedOperationException("Unknown operation type: " + type);
			}
		} else {
			return super.calculateSparseLocalDerivative(a, b, type);
		}
	}

	private SparseLocalDerivative multiplyWithSkipFactor(SparseLocalDerivative deriv) {
		final Iterator<SparseDeriativeEntry> iterator = deriv.iterator();
		while (iterator.hasNext()) {
			final SparseDeriativeEntry next = iterator.next();
			next.setValue(skipFactor * next.getValue());
		}
		return deriv;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int hashCode() {
		int hash = 7;
		hash = 47 * hash + (this.dynamic ? 1 : 0);
		hash = 47 * hash + Objects.hashCode(this.deleteComparison);
		hash = 47 * hash + Objects.hashCode(this.insertComparison);
		hash = 47 * hash + Objects.hashCode(this.skipDeleteComparison);
		hash = 47 * hash + Objects.hashCode(this.skipInsertComparison);
		hash = 47 * hash + super.hashCode();
		return hash;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final ComparisonBasedSkipExtendedComparator<?> other = (ComparisonBasedSkipExtendedComparator<?>) obj;
		if (this.dynamic != other.dynamic) {
			return false;
		}
		if (!Objects.equals(this.deleteComparison, other.deleteComparison)) {
			return false;
		}
		if (!Objects.equals(this.insertComparison, other.insertComparison)) {
			return false;
		}
		if (!Objects.equals(this.skipDeleteComparison, other.skipDeleteComparison)) {
			return false;
		}
		if (!Objects.equals(this.skipInsertComparison, other.skipInsertComparison)) {
			return false;
		}
		if (!super.equals(obj)) {
			return false;
		}
		return true;
	}

}
