/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.comparators;

import de.unibi.citec.tcs.alignment.OperationType;
import de.unibi.citec.tcs.alignment.sequence.ValueType;
import de.unibi.citec.tcs.alignment.sequence.VectorialValue;

/**
 * implements halved gauss-kernel k:
 *
 * k = exp(-1/2sigma^2 * |x-y|^2) /2, with |.| beeing the weighted euclidian
 * distance
 *
 * k is positive definite
 *
 * @author Georg Zentgraf
 */
public class DTWKernelComparator implements Comparator<VectorialValue>,
	DerivableComparator<VectorialValue, double[]> {

	// weights for weighted euclidian distance -> these are the derivable parameters
	final private double[] localWeights;
	// standard deviation for gauss kernel
	private double sigma = 1;

	public DTWKernelComparator(double[] localWeights) {
		this.localWeights = localWeights;
	}

	public DTWKernelComparator(double[] localWeights, double sigma) {
		this.localWeights = localWeights;
		this.sigma = sigma;
	}

	public ValueType getType() {
		return ValueType.VECTOR;
	}

	/**
	 *
	 * @return standard deviation of gauss kernel.
	 */
	public double getSigma() {
		return this.sigma;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double compare(VectorialValue a, VectorialValue b) {
		if (a.getVector().length != localWeights.length || 
                b.getVector().length != localWeights.length ) {
			throw new RuntimeException("Local weight vector and local vectors have diffrent length!");
		}
		double gaussKernel = calc_gauss_kernel(a.getVector(), b.getVector());
		return gaussKernel / 2;
	}

	/**
	 * {@inheritDoc }
	 */
//	@Override
//	public double[] calculateDerivative(AlignmentDerivativeAlgorithm derivAlgo) {
//		final int size = localWeights.length;
//		final double[] devrivativeVector = new double[size];
//		for (int i = 0; i < size; i++) {
//			devrivativeVector[i] = derivAlgo.calculateDerivative(this, i);
//		}
//		return devrivativeVector;
//	}

	/**
	 * TODO: think about opterationType
	 *
	 * {@inheritDoc }
	 */
	@Override
	public double calculateLocalDerivative(int paramIdent, VectorialValue a, VectorialValue b, OperationType type) {
		double[] aV = a.getVector(), bV = b.getVector();
		double numerator = -calc_gauss_kernel(aV, bV) * (aV[paramIdent] - bV[paramIdent]) * (aV[paramIdent] - bV[paramIdent]);
		double denominator = 4 * this.sigma * this.sigma;
		return numerator / denominator;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public double[] transformToResult(double[] derivatives) {
		return derivatives;
	}

	@Override
	public Class<double[]> getResultClass() {
		return double[].class;
	}

	@Override
	public int getNumberOfParameters() {
		return localWeights.length;
	}

//-----------------------------------------------------------------------------
	/**
	 * Returns gauss kernel value for vector a and vector b.
	 *
	 * @param a a double vector.
	 * @param b a double vector.
	 *
	 * @return gauss kernel value for vector a and vector b.
	 */
	private double calc_gauss_kernel(double[] a, double[] b) {
		double dist = weighted_squared_dist(a, b);
		return Math.exp(-1 / (2 * this.sigma * this.sigma) * dist);
	}

	/**
	 * Returns the squared weighted euclidian distance between vector a and
	 * vector b.
	 *
	 * @param a a double vector.
	 * @param b a double vector.
	 *
	 * @return squared weighted euclidian distance between vector a and vector
	 * b.
	 */
	private double weighted_squared_dist(double[] a, double[] b) {
		double localdist, d = 0;
		for (int i = 0; i < a.length; i++) {
			localdist = a[i] - b[i];
			d += localWeights[i] * localdist * localdist;
		}
		return d;
	}

}
