/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.comparators;

import de.unibi.citec.tcs.alignment.OperationType;
import de.unibi.citec.tcs.alignment.sequence.Value;

/**
 * Some derivable comparators have the special property that in each alignment
 * step only a small subset of parameters is needed. This allows for more
 * runtime efficiency in derivation calculation.
 *
 * For example: The TrivialEditComparator has distinct parameters for match,
 * mismatch, gap and skip cost. The parameters only come into play if such an
 * operation was actually used. Thus we can omit calculations of the parameters
 * for other operations at that step.
 *
 * @param <X> The value that can be compared by this class.
 * @param <Y> An intuitive format for the derivatives of all parameters of this
 * comparator.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public interface SparseDerivableComparator<X extends Value, Y>
	extends DerivableComparator<X, Y> {

	/**
	 * This method shall return a sparse representation of the derivative for
	 * all parameters that have non-zero derivatives in this step of the
	 * alignment.
	 *
	 * @param a the value that has been compared with b.
	 * @param b the value that has been compared with a.
	 * @param type the operation type that has actually been used in the
	 * comparison.
	 *
	 * @return a sparse representation of the derivative for all parameters that
	 * have non-zero derivatives in this step of the alignment.
	 */
	public SparseLocalDerivative calculateSparseLocalDerivative(
		X a, X b, OperationType type);

}
