/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.sequence;

/**
 * This is a special value class that is restricted to a finite list of values
 * defined by an alphabet.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class SymbolicValue extends AbstractValue {

	private final Alphabet alphabet;
	private String symbol;

	public SymbolicValue(Alphabet alphabet) {
		super(ValueType.SYMBOLIC);
		this.alphabet = alphabet;
	}

	public SymbolicValue(Alphabet alphabet, String symbol) {
		super(ValueType.SYMBOLIC);
		this.alphabet = alphabet;
		this.symbol = symbol;
		if (!alphabet.hasSymbol(symbol)) {
			throw new UnsupportedOperationException(
					"The given alphabet does not support the given symbol!");
		}
	}

	/**
	 * Returns the Alphabet that defines which symbols are allowed for this
	 * keyword.
	 *
	 * @return the Alphabet that defines which symbols are allowed for this
	 * keyword.
	 */
	public Alphabet getAlphabet() {
		return alphabet;
	}

	/**
	 * Returns this actual symbol.
	 *
	 * @return this actual symbol.
	 */
	public String getSymbol() {
		return symbol;
	}

	/**
	 * Sets the actual symbol.
	 *
	 * @param symbol the actual symbol.
	 */
	public void setSymbol(String symbol) {
		if (alphabet.hasKeyword(symbol)) {
			this.symbol = symbol;
		} else {
			throw new UnsupportedOperationException(
					"The given alphabet does not support the given symbol!");
		}
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public String toString() {
		return symbol.toString();
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public int hashCode() {
		int hash = 7;
		hash = 97 * hash + (this.symbol != null ? this.symbol.hashCode() : 0);
		return hash;
	}

	/**
	 * {@inheritDoc }
	 */
	@Override
	public boolean equals(Object obj) {
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final SymbolicValue other = (SymbolicValue) obj;
		if ((this.symbol == null) ? (other.symbol != null) : !this.symbol.equals(other.symbol)) {
			return false;
		}
		return true;
	}
}
