/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.unibi.citec.tcs.alignment.visualization;
//package de.unibi.citec.fit.dfgvis;

import static de.unibi.citec.tcs.alignment.visualization.Visualization.ERROR_FILE_PERMISSION;
//import static de.unibi.citec.fit.dfgvis.Visualization.ERROR_FILE_PERMISSION;
import de.unibi.citec.tcs.alignment.AlignmentPath;
import de.unibi.citec.tcs.alignment.Operation;
import java.io.*;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.*;

/**
 *
 * @author ckrueger
 */
public class Utils {
    /**
     * Formats a given Number by adding zeros at the beginning.
     * @param number Number to be formatted.
     * @param digits Length the number shall have.
     * @return Formatted number.
     */
    public static String formatNumber(int number, int digits) {
        /** source: http://www.rgagnon.com/javadetails/java-0021.html
         * we create a format :
         *  %% : %  the first % is to escape the second %
         *  0  : 0  zero character
         *  %d :    how many '0' will be added (that many, such that the 
         *              complete number is digits long)
         *  d  : d  the number to format
         */
        String format = String.format("%%0%dd", digits);
        return String.format(format, number);
    }
    
    /**
     * Finds a file at its paths, reads it in and returns its 
     * content as a list of Strings.
     * Every entry represents a line in the file. Order is preserved.
     * @param path The file's absolut path.
     * @return An ArrayList of Strings.
     */
    public static ArrayList<String> readFileFromPath(String path) {
        final File file = new File(path);
        return readFile(file);
    }
    
    /**
     * Reads a File and returns its content as a list of Strings.
     * Every entry represents a line in the file. Order is preserved.
     * @param file The file.
     * @return An ArrayList of Strings.
     */
    public static ArrayList<String> readFile(File file) {
        final ArrayList<String> text = new ArrayList<String>();
        BufferedReader stream = null;
        FileReader fileReader = null;
        String path = file.getAbsolutePath();
        try {
            if (!(file.exists() && file.canRead())) 
                throw new RuntimeException(ERROR_FILE_PERMISSION + ": " + path);

            fileReader = new FileReader(file);
            stream = new BufferedReader(fileReader);

            String line;
            while ((line = stream.readLine()) != null) {
                text.add(line);
            }
        } catch (IOException ioEx) {
            throw new RuntimeException("Reading file " + path + "was interrupted."
                    + "\n IOException:\n" + ioEx.getMessage());
        } finally {
            try {
                if (stream != null) {
                    stream.close();
                }
                if (fileReader != null) {
                    fileReader.close();
                }
            } catch (IOException ioEx) {
                throw new RuntimeException("Closing file " + path + "was interrupted."
                    + "\n IOException:\n" + ioEx.getMessage());
            }
        }
        return text;
    }
    
    /**
     * Writes the given text to a file.
     * @param path The file's absolut path.
     * @param content The text, which should be written to the file.
     */
    public static void writeFile(String path, String content) {
        BufferedWriter stream = null;
        FileWriter fileWriter = null;
        final File file = new File(path);
        try {
            fileWriter = new FileWriter(file);
            stream = new BufferedWriter(fileWriter);

            stream.write(content);
        } catch (IOException ioEx) {
            throw new RuntimeException("Writing file " + path + "was interrupted."
                    + "\n IOException:\n" + ioEx.getMessage());
        } finally {
            try {
                if (stream != null) {
                    stream.close();
                }
                if (fileWriter != null) {
                    fileWriter.close();
                }
            } catch (IOException ioEx) {
                throw new RuntimeException("Closing file " + path + "was interrupted."
                    + "\n IOException:\n" + ioEx.getMessage());
            }
        }
    }

    /**
     * Calculates a color scheme for all keys over all alignments.
     * Every keyword gets its own mapping of double-values to rgb-colors gained 
     * by looking over all alignments.
     * @param alignments All alignments that should be included.
     * @return A mapping of all occuring doubles to an calculated rgb-value for 
     * every occuring key.
     */
    public static Map<String, Map<Double, Integer[]> > rgbColorsForKeys(
            final AlignmentPath[][] alignments) {
        
        final ArrayList<String> keys = new ArrayList<String>();
        final ArrayList<Double[]> minAndMax = new ArrayList<Double[]>();
        final Double[] tmpDou = {Double.POSITIVE_INFINITY, Double.NEGATIVE_INFINITY};
        
        //final coloring scheme
        final Map<String, Map<Double, Integer[]> > distColorForKeys 
                = new HashMap<String, Map<Double, Integer[]> >();
        //tmp, only to collect all possible double values
        final Map<String, Set<Double> > scoresOfAllOperations 
                = new HashMap<String, Set<Double> >();
        
        AlignmentPath testAlignment = alignments[0][0];
        both:
        for (int i = 0; i < alignments.length; i++) {
            for (int j = 0; j < alignments.length; j++) {
                if (alignments[i][j] != null) {
                    testAlignment = alignments[i][j];
                    break both;
                }
            }
        }
        for (final String str : testAlignment.getSpecificaton().getKeywords()) {
            keys.add(str);
            minAndMax.add(tmpDou);
            scoresOfAllOperations.put(str, new HashSet<Double>());
        }
        
        
        final DecimalFormatSymbols otherSymbols = new DecimalFormatSymbols(Locale.getDefault());
        otherSymbols.setDecimalSeparator('.');
        otherSymbols.setGroupingSeparator(',');
        final DecimalFormat df = new DecimalFormat( "#,###,###,##0.0000", otherSymbols);
        
        for (int i = 0; i < alignments.length; i++) {   
            for (int j = i; j < alignments.length; j++) {
                ArrayList<Operation> ops;
                if (alignments[i][j] != null) {
                    ops = alignments[i][j].getOperations();
                } else if (alignments[j][i] != null) {
                    ops = alignments[j][i].getOperations();
                } else {
                    continue;
                }
                
                for (final Operation op : ops) {
                    for (final String str : keys) {
                        final int origIdx = keys.indexOf(str);
                        final double score = Double.parseDouble(df.format(
                                op.getComparatorDistances()[origIdx]));
                        
                        //returns true if str is new to the HashMap
                        if (scoresOfAllOperations.get(str).add(score)) {
                            if (score < minAndMax.get(origIdx)[0]) 
                                minAndMax.get(origIdx)[0] = score;
                            else if (score > minAndMax.get(origIdx)[1])
                                minAndMax.get(origIdx)[1] = score;
                        }
                    }
                }
            } 
        }
        for (final String str : scoresOfAllOperations.keySet()) {
            distColorForKeys.put(str, realValuesToRgbColors(
                    scoresOfAllOperations.get(str), 
                    minAndMax.get(keys.indexOf(str))[1], 
                    minAndMax.get(keys.indexOf(str))[0]));
        }
        return distColorForKeys;
    }
    
    /**
     * Calculates a color scheme for all alignment scores (for form.html).
     * @param alignments Alignment matrix, containing the scores.
     * @return A map, where for every score that appears, suitable rgb values exist.
     */
    public static Map<Double, Integer[]> rgbColorsForAlignments(
            final AlignmentPath[][] alignments) {
        
        final int alignLength = alignments.length;
        final Set<Double> scores = new HashSet<Double>();
        double score;
        double max = Double.NEGATIVE_INFINITY;
        double min = Double.POSITIVE_INFINITY;      
                
        final DecimalFormatSymbols otherSymbols = new DecimalFormatSymbols(Locale.getDefault());
        otherSymbols.setDecimalSeparator('.');
        otherSymbols.setGroupingSeparator(',');
        final DecimalFormat df = new DecimalFormat( "#,###,###,##0.00", otherSymbols);
        
        for (int i = 0; i < alignLength; i++) {
            for (int j = 0; j < alignLength; j++) {
                if (alignments[i][j] == null) continue;
                
                score = Double.parseDouble(df.format(alignments[i][j].getScore()));
                
                if (scores.add(score)) {
                    if (score > max) max = score;
                    if (min > score) min = score;
                }
            }
        }
        
        if (min == max) {
            min = 0;
            max = 1;
        }
        return realValuesToRgbColors(scores, max, min);
    }
    
    /**
     * Calculates a color scheme for all operation  scores in an alignment.
     * @param operations Needed to retrieve the scores.
     * @return A map, where for every score that appears, a suitable rgb values exist.
     */
    public static Map<Double, Integer[]> rgbColorsForOperations(
            final ArrayList<Operation> operations) {
        
        final int alignLength = operations.size();
        final Set<Double> scores = new HashSet<Double>();
        double score;
        double max = Double.NEGATIVE_INFINITY;
        double min = Double.POSITIVE_INFINITY;      
        
        final DecimalFormatSymbols otherSymbols = new DecimalFormatSymbols(Locale.getDefault());
        otherSymbols.setDecimalSeparator('.');
        otherSymbols.setGroupingSeparator(',');
        final DecimalFormat df = new DecimalFormat( "#,###,###,##0.0000", otherSymbols);
        
        for (int i = 0; i < alignLength; i++) {
            score = Double.parseDouble(df.format(
                    operations.get(i).getWeightedLocalCost()));
            if (scores.add(score)) {
                if (score > max) max = score;
                if (min > score) min = score;
            }
        }
        
        if (min == max) {
            min = 0;
            max = 1;
        }
        return realValuesToRgbColors(scores, max, min);
    }
    
    /**
     * Calculates a coloring scheme from dark green to yellow from given values. 
     * Maps double values to rgb values, represented by a list of 
     * three values (red, green, blue). 
     * @param scores 
     * @param max Needed for normalization. Maximum value in scores.
     * @param min Needed for normalization. Minimal value in scores.
     * @return A map, where for every score that appears, a suitable rgb values exist.
     */
    public static Map<Double, Integer[]> realValuesToRgbColors(final Set<Double> scores,
            final double max, final double min) {

        final Map<Double, Integer[]> distColor = new HashMap<Double, Integer[]>();
        double normScore;
        //red [0,255] green [128,255] blue [102,102]
        Integer red, green;
        final Integer blue = 102;
        for(final Double dou : scores) {
            //normalize
            normScore = (dou - min) / (max - min);
            
            //(int)(long) is a cast from long to Integer
            red = (int)(long)Math.round(255 * normScore);
            green = (int)(long)Math.round(127.5 + (127.5 * normScore));
            
            final Integer[] tmpInt = {red, green, blue};
            distColor.put(dou, tmpInt);
        }

        return distColor;
    }
}
