/* 
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment.wrappers;

import de.unibi.citec.tcs.alignment.AlignmentAlgorithm;
import de.unibi.citec.tcs.alignment.AlignmentDerivativeAlgorithm;
import de.unibi.citec.tcs.alignment.AlignmentSpecification;
import de.unibi.citec.tcs.alignment.ParallelDerivativeEngine;
import de.unibi.citec.tcs.alignment.ParallelProcessingEngine;
import de.unibi.citec.tcs.alignment.comparators.Comparator;
import de.unibi.citec.tcs.alignment.comparators.ReplacementComparator;
import de.unibi.citec.tcs.alignment.comparators.ReplacementCosts;
import de.unibi.citec.tcs.alignment.sequence.Alphabet;
import de.unibi.citec.tcs.alignment.sequence.NodeSpecification;
import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.sequence.SymbolicKeywordSpecification;
import java.util.Arrays;
import java.util.Collection;
import java.util.TreeSet;

/**
 * This is a simple wrapper to support an easier interface for the simple
 * StringEditDistance problem.
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class StringEditDistance {

	/**
	 * Transforms the given strings to the TCSAlignmentToolbox Sequence format.
	 * This method assumes that the input strings contain all symbols that are
	 * relevant to your dataspace. If that is not the case you should explicitly
	 * specify your Alphabet.
	 *
	 * @param strings your data. The input format can be either a plain string,
	 * where each character is interpreted as a distinct symbol or a string of
	 * symbols separated by a vertical bar '|' character.
	 * @return the TCSAlignmentToolbox Sequence representation of your data.
	 */
	public static Sequence[] toSequences(String[] strings) {
		return toSequences(Arrays.asList(strings));
	}

	/**
	 * Transforms the given strings to the TCSAlignmentToolbox Sequence format.
	 * This method assumes that the input strings contain all symbols that are
	 * relevant to your dataspace. If that is not the case you should explicitly
	 * specify your Alphabet.
	 *
	 * @param strings your data. The input format can be either a plain string,
	 * where each character is interpreted as a distinct symbol or a string of
	 * symbols separated by a vertical bar '|' character.
	 * @return the TCSAlignmentToolbox Sequence representation of your data.
	 */
	public static Sequence[] toSequences(Collection<String> strings) {
		final TreeSet<String> symbols = new TreeSet<>();
		String delimEncodingStr = null;
		for (final String string : strings) {
			if (string.contains("|")) {
				if (delimEncodingStr == null) {
					delimEncodingStr = string;
				}
				//if we have a delimiter, interpret everything in between as symbols.
				final String[] split = string.split("\\|");
				for (final String symbol : split) {
					symbols.add(symbol);
				}
			} else {
				if (delimEncodingStr != null) {
					throw new UnsupportedOperationException(delimEncodingStr
							+ "\n\nwas interpreted as |-delimited symbol string, but\n\n"
							+ string
							+ "\n\ncontained no delimiter. Please ensure a consistent encoding!");
				}
				final char[] chars = string.toCharArray();
				for (int i = 0; i < chars.length; i++) {
					symbols.add(new String(new char[]{chars[i]}));
				}
			}
		}
		final Alphabet alphabet = new Alphabet(symbols.toArray(new String[0]));
		return toSequences(strings, alphabet);
	}

	/**
	 * Transforms the given strings to the TCSAlignmentToolbox Sequence format
	 * using the given alphabet. If the given alphabet does not specify a symbol
	 * that is contained in your input, an UnsupportedOperationIndex is thrown.
	 *
	 * @param strings your data. The input format can be either a plain string,
	 * where each character is interpreted as a distinct symbol or a string of
	 * symbols separated by a vertical bar '|' character.
	 * @param alphabet the alphabet containing all symbols of your data space.
	 * @return the TCSAlignmentToolbox Sequence representation of your data.
	 */
	public static Sequence[] toSequences(String[] strings, Alphabet alphabet) {
		return toSequences(Arrays.asList(strings), alphabet);
	}

	/**
	 * Transforms the given strings to the TCSAlignmentToolbox Sequence format
	 * using the given alphabet. If the given alphabet does not specify a symbol
	 * that is contained in your input, an UnsupportedOperationIndex is thrown.
	 *
	 * @param strings your data. The input format can be either a plain string,
	 * where each character is interpreted as a distinct symbol or a string of
	 * symbols separated by a vertical bar '|' character.
	 * @param alphabet the alphabet containing all symbols of your data space.
	 * @return the TCSAlignmentToolbox Sequence representation of your data.
	 */
	public static Sequence[] toSequences(Collection<String> strings, Alphabet alphabet) {
		final Sequence[] out = new Sequence[strings.size()];
		int s = 0;
		for (final String string : strings) {
			out[s] = new Sequence(string, alphabet);
			s++;
		}
		return out;
	}

	/**
	 * Sets up a default AlignmentSpecification for the simple
	 * StringEditDistance problem. The scoring scheme is as follows:
	 * <ul><li>replacements of equal symbols cost nothing.</li>
	 * <li>replacements of unequal symbols cost 1.</li>
	 * <li>deletions and insertions cost 1.</li></ul>
	 * This is equivalent to the Damerau-Levenshtein distance without swaps.
	 *
	 * @param dataSpace the sequences of your dataspace.
	 * @return an AlignmentSpecification with default costs.
	 */
	public static AlignmentSpecification setUpSpecification(Sequence[] dataSpace) {
		if (dataSpace == null || dataSpace.length == 0) {
			throw new UnsupportedOperationException("The given data space is empty!");
		}
		//check for NodeSpecification consistency.
		final NodeSpecification nodeSpec = dataSpace[0].getNodeSpecficiation();
		for (int s = 1; s < dataSpace.length; s++) {
			if (!nodeSpec.equals(dataSpace[s].getNodeSpecficiation())) {
				throw new UnsupportedOperationException("The NodeSpecifications in the given data space are not consistent!");
			}
		}
		//retrieve Alphabet.
		final Alphabet alphabet = ((SymbolicKeywordSpecification) nodeSpec.
				getKeywordSpecification(0)).getAlphabet();
		final ReplacementComparator comp = new ReplacementComparator(
				new ReplacementCosts(alphabet, true, true));
		//construct AlignmentSpecification.
		return new AlignmentSpecification(nodeSpec,
				new String[]{Sequence.DEFAULTKEYWORD},
				new Comparator[]{comp});
	}

	/**
	 * Sets up an AlignmentSpecification for the simple
	 * StringEditDistance problem.
	 *
	 * @param dataSpace the sequences of your dataspace.
	 * @param matchCost the cost for aligning two equal symbols.
	 * @param mismatchCost the cost for aligning two unequal symbols.
	 * @param gapCost the cost for deleting or inserting a symbol.
	 * @return an AlignmentSpecification with your defined scoring scheme.
	 */
	public static AlignmentSpecification setUpSpecification(Sequence[] dataSpace,
			double matchCost, double mismatchCost, double gapCost) {
		if (dataSpace == null || dataSpace.length == 0) {
			throw new UnsupportedOperationException("The given data space is empty!");
		}
		//check for NodeSpecification consistency.
		final NodeSpecification nodeSpec = dataSpace[0].getNodeSpecficiation();
		for (int s = 1; s < dataSpace.length; s++) {
			if (!nodeSpec.equals(dataSpace[s].getNodeSpecficiation())) {
				throw new UnsupportedOperationException("The NodeSpecifications in the given data space are not consistent!");
			}
		}
		//retrieve Alphabet.
		final Alphabet alphabet = ((SymbolicKeywordSpecification) nodeSpec.
				getKeywordSpecification(0)).getAlphabet();
		//construct Comparator.
		final double[][] costMatrix = new double[alphabet.size() + 2][alphabet.size() + 2];
		for (int i = 0; i < alphabet.size() + 2; i++) {
			for (int j = 0; j < alphabet.size() + 2; j++) {
				if (i >= alphabet.size() || j >= alphabet.size()) {
					costMatrix[i][j] = gapCost;
				} else {
					if (i == j) {
						costMatrix[i][j] = matchCost;
					} else {
						costMatrix[i][j] = mismatchCost;
					}
				}
			}
		}
		final ReplacementCosts costs = new ReplacementCosts(alphabet, costMatrix);
		final ReplacementComparator comp = new ReplacementComparator(costs);
		//construct AlignmentSpecification.
		return new AlignmentSpecification(nodeSpec,
				new String[]{Sequence.DEFAULTKEYWORD},
				new Comparator[]{comp});
	}

	/**
	 * Sets up an AlignmentSpecification for the simple
	 * StringEditDistance problem.
	 *
	 * @param dataSpace the sequences of your dataspace.
	 * @param scoringScheme a scoring scheme that defines individual replacement
	 * and gap costs for each symbol. More formally: For each combination of
	 * symbols a,b in Sigma u {_,skip} you have to specify
	 *
	 * lambda(a,b), where lambda(a,b) is stored at point [i][j] in the matrix,
	 * where i is the index of a in the alphabet and j is the index of b in the
	 * alphabet (see also the Alphabet documentation).
	 *
	 * deletion costs
	 *
	 * lambda(a,_)
	 *
	 * are stored at point [i][|Sigma|] and insertion costs
	 *
	 * lambda(_,b)
	 * @return an AlignmentSpecification with your defined scoring scheme.
	 */
	public static AlignmentSpecification setUpSpecification(Sequence[] dataSpace,
			double[][] scoringScheme) {
		if (dataSpace == null || dataSpace.length == 0) {
			throw new UnsupportedOperationException("The given data space is empty!");
		}
		//check for NodeSpecification consistency.
		final NodeSpecification nodeSpec = dataSpace[0].getNodeSpecficiation();
		for (int s = 1; s < dataSpace.length; s++) {
			if (!nodeSpec.equals(dataSpace[s].getNodeSpecficiation())) {
				throw new UnsupportedOperationException("The NodeSpecifications in the given data space are not consistent!");
			}
		}
		//retrieve Alphabet.
		final Alphabet alphabet = ((SymbolicKeywordSpecification) nodeSpec.
				getKeywordSpecification(0)).getAlphabet();
		final ReplacementCosts costs = new ReplacementCosts(alphabet, scoringScheme);
		final ReplacementComparator comp = new ReplacementComparator(costs);
		//construct AlignmentSpecification.
		return new AlignmentSpecification(nodeSpec,
				new String[]{Sequence.DEFAULTKEYWORD},
				new Comparator[]{comp});
	}

	/**
	 * Calculates the alignment derivative between the two given input sequences
	 * using the given algorithm.
	 *
	 * @param algorithm an algorithm that returns enough information to
	 * calculate the alignment.
	 * @param a a sequence.
	 * @param b a sequence.
	 * @return the alignment derivative between the two given input sequences.
	 * The format is that of a cost matrix: It specifies the derivative for each
	 * possible combination of symbols and gaps in the data space. The order of
	 * symbols if given by the alphabet.
	 */
	public static double[][] calculateDerivative(
			AlignmentAlgorithm<? extends AlignmentDerivativeAlgorithm> algorithm,
			Sequence a, Sequence b) {
		final AlignmentDerivativeAlgorithm result
				= algorithm.calculateAlignment(a, b);
		//retrieve comparator.
		final ReplacementComparator comp
				= (ReplacementComparator) algorithm.getSpecification().getComparator(0);
		//return result.
		return result.calculateParameterDerivative(comp, Sequence.DEFAULTKEYWORD);
	}

	/**
	 * Calculates the pairwise alignment derivative between all given input
	 * sequences using the given algorithm. The pairwise derivatives are
	 * calculated in parallel for faster computation.
	 *
	 * @param algorithm an algorithm that returns enough information to
	 * calculate the alignment.
	 * @param dataSpace an array of sequences.
	 * @param threadNum the number of threads that shall be used for parallel
	 * computation.
	 * @return the alignment derivative between the two given input sequences.
	 * The format is that of a cost matrix: It specifies the derivative for each
	 * possible combination of symbols and gaps in the data space. The order of
	 * symbols if given by the alphabet.
	 */
	public static <R extends AlignmentDerivativeAlgorithm> double[][][][] calculateDerivatives(
			AlignmentAlgorithm<R> algorithm,
			Sequence[] dataSpace, int threadNum) {
		//retrieve comparator.
		final ReplacementComparator comp
				= (ReplacementComparator) algorithm.getSpecification().getComparator(0);
		//calculate alignments
		final int N = dataSpace.length;
		final ParallelProcessingEngine<R> procEngine
				= new ParallelProcessingEngine<R>(algorithm, dataSpace);
		procEngine.setNumberOfThreads(threadNum);
		procEngine.calculate();
		final R[][] resultMatrix = procEngine.getResultMatrix();
		//calculate derivatives
		final ParallelDerivativeEngine derivEngine
				= new ParallelDerivativeEngine(resultMatrix);
		derivEngine.setNumberOfThreads(threadNum);
		final double[][][] comparatorDerivatives
				= derivEngine.calculateComparatorDerivatives(
						comp, Sequence.DEFAULTKEYWORD);
		//rearrange result.
		final double[][][][] out = new double[N][N][][];
		int k = 0;
		for (int i = 0; i < N; i++) {
			for (int j = 0; j < N; j++) {
				out[i][j] = comparatorDerivatives[k];
				k++;
			}
		}
		return out;
	}
}
