/*
 * TCS Alignment Toolbox
 * 
 * Copyright (C) 2013, 2014
 * Benjamin Paaßen, Charlie Krüger, Georg Zentgraf
 * AG Theoretical Computer Science
 * Centre of Excellence Cognitive Interaction Technology (CITEC)
 * University of Bielefeld
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package de.unibi.citec.tcs.alignment;

import de.unibi.citec.tcs.alignment.sequence.Sequence;
import de.unibi.citec.tcs.alignment.wrappers.VectorialSequences;
import java.util.Random;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Test;
import static org.junit.Assert.*;

/**
 *
 * @author Benjamin Paassen - bpaassen@techfak.uni-bielefeld.de
 */
public class StrictDTWScoreAlgorithmTest {

	public StrictDTWScoreAlgorithmTest() {
	}

	@BeforeClass
	public static void setUpClass() {
	}

	@AfterClass
	public static void tearDownClass() {
	}

	@Before
	public void setUp() {
	}

	@After
	public void tearDown() {
	}

	/**
	 * Test of DTW calculation.
	 */
	@Test
	public void testCalculate() {
		final int numSamples = 100;
		final double f1 = 0.1;
		final double f2 = 0.07;
		//create two sine waves with different frequencies.
		final double[][] waves = new double[2][numSamples];
		for (int i = 0; i < numSamples; i++) {
			waves[0][i] = Math.sin(Math.PI * f1 * i);
			waves[1][i] = Math.sin(Math.PI * f2 * i);
		}
		//transform them to sequences.
		final Sequence[] seqs = VectorialSequences.toSequences(waves);
		//create AlignmentSpecification.
		final AlignmentSpecification spec = VectorialSequences.setUpSpecification(seqs);
		//set up algorithm.
		final StrictDTWScoreAlgorithm instance = new StrictDTWScoreAlgorithm(spec);

		//calculate the full distance matrix.
		final ParallelProcessingEngine<Double> parallelProc
				= new ParallelProcessingEngine<>(instance, seqs);
		parallelProc.setVerbose(false);
		parallelProc.calculate();
		final double[][] scoreMatrix = parallelProc.getScoreMatrix();

		//assert zero distance on diagonal.
		for (int s = 0; s < waves.length; s++) {
			assertEquals(0, scoreMatrix[s][s], 0);
		}
		//assert symmetry
		assertEquals(scoreMatrix[0][1], scoreMatrix[1][0], 0);
		//also the distance between both sequences should be fairly low, because
		//DTW should be able to account for different frequencies.
		assertTrue(scoreMatrix[0][1] >= 0);
		assertTrue(scoreMatrix[0][1] < 0.2);
	}

	/**
	 * Test of DTW performance.
	 */
	@Test
	public void testPerformance() {
		final int seqLength = 1000;
		final int numFeatures = 15;

		//Create data set
		final double[][][] waves = new double[2][seqLength][numFeatures];
		for (int i = 0; i < seqLength; i++) {
			for (int k = 0; k < numFeatures; k++) {
				final double f = 1. / (double) (k + 1);
				waves[0][i][k] = Math.sin(Math.PI * f * i);
				//the second multi-dimensional wave should have a slight phase shift.
				waves[1][i][k] = Math.sin(Math.PI * f * (i + 1));
			}
		}

		//set up a random weighting (with some zero weights).
		final Random rand = new Random();
		final String[] labels = new String[numFeatures];
		final double[] weights = new double[numFeatures];
		double normalization = 0;
		for (int k = 0; k < numFeatures; k++) {
			labels[k] = Integer.toString(k);
			weights[k] = rand.nextGaussian();
			if(weights[k] < 0){
				weights[k] = 0;
			}
			normalization += weights[k];
		}

		for (int k = 0; k < numFeatures; k++) {
			weights[k] /= normalization;
		}

		//transform them to sequences.
		final Sequence[] seqs = VectorialSequences.toSequences(waves, labels);
		//create AlignmentSpecification.
		final AlignmentSpecification spec = VectorialSequences.setUpSpecification(seqs, weights);
		//set up algorithm.
		final StrictDTWScoreAlgorithm instance = new StrictDTWScoreAlgorithm(spec);
		//look at runtime.
		final long start = System.nanoTime();
		final Double score = instance.calculateAlignment(seqs[0], seqs[1]);
		final long time = (System.nanoTime() - start) / (long) 1E6;
		assertTrue(score >= 0);
		assertTrue(score <= 0.01);
	}
}
